use core::arch::asm;

/// IO base address for master PIC
const PIC_A: u32 = 0x20;
const PIC_A_COMMAND: u32 = PIC_A;
const PIC_A_DATA: u32 = PIC_A + 1;

/// IO base address for slave PIC
const PIC_B: u32 = 0xA0;
const PIC_B_COMMAND: u32 = PIC_B;
const PIC_B_DATA: u32 = PIC_B + 1;

/// PIC `EOI` command
const CMD_EOI: u8 = 0x20;

unsafe fn outb(port: u32, value: u8) {
	asm!(
		"out dx, al",
		in("dx") port,
		in("al") value,
	);
}

/// A driver for the PIC
///
/// Reference:
/// - https://wiki.osdev.org/8259_PIC
/// - https://os.phil-opp.com/hardware-interrupts
pub struct PICDriver {
	offset_pic_a: u8,
	offset_pic_b: u8,
}

impl PICDriver {
	pub const fn new(offset_pic_a: u8, offset_pic_b: u8) -> Self {
		Self {
			offset_pic_a,
			offset_pic_b,
		}
	}

	fn send_a_cmd(&self, cmd: u8) {
		unsafe { outb(PIC_A_COMMAND, cmd) }
	}

	fn send_a_data(&self, cmd: u8) {
		unsafe { outb(PIC_A_DATA, cmd) }
	}

	fn send_b_cmd(&self, cmd: u8) {
		unsafe { outb(PIC_B_COMMAND, cmd) }
	}

	fn send_b_data(&self, cmd: u8) {
		unsafe { outb(PIC_B_DATA, cmd) }
	}

	pub fn send_eoi(&self, irq: u8) {
		if irq > 8 {
			self.send_b_cmd(CMD_EOI);
		}

		self.send_a_cmd(CMD_EOI);
	}

	pub fn init(&mut self) {
		const ICW1_ICW4: u8 = 0x01; /* Indicates that ICW4 will be present */
		const ICW1_INIT: u8 = 0x10; /* Initialization - required! */
		const ICW4_8086: u8 = 0x01; /* 8086/88 (MCS-80/85) mode */

		self.send_a_cmd(ICW1_INIT | ICW1_ICW4);
		self.send_b_cmd(ICW1_INIT | ICW1_ICW4);

		self.send_a_data(self.offset_pic_a); // ICW2: Master PIC vector offset
		self.send_b_data(self.offset_pic_b); // ICW2: Slave PIC vector offset

		self.send_a_data(4); // ICW3: tell Master PIC that there is a slave PIC at IRQ2 (0000 0100)
		self.send_b_data(2); // ICW3: tell Slave PIC its cascade identity (0000 0010)

		// ICW4: have the PICs use 8086 mode (and not 8080 mode)
		self.send_a_data(ICW4_8086);
		self.send_b_data(ICW4_8086);

		// Unmask both PICs
		self.send_a_data(0);
		self.send_b_data(0);
	}
}
