use core::{
	fmt::{self},
	marker::PhantomData,
};

use super::{HandlerFuncType, VirtAddr};

/// An Interrupt Descriptor Table entry.
///
/// The generic parameter is some [`HandlerFuncType`], depending on the interrupt vector.
///
/// For reference, see https://wiki.osdev.org/Interrupt_Descriptor_Table#Gate_Descriptor
#[derive(Clone, Copy)]
#[repr(C)]
pub struct Entry<F> {
	phantom: PhantomData<F>,

	/// Low 16 bits of the 32-bit ISR pointer
	pointer_low: u16,

	/// A segment selector that points to a valid code segment in the GDT
	segment: u16,

	/// Always zero
	reserved: u8,

	/// Interrupt options.
	///
	/// Bits 0-3: gate type. One of:
	/// - 0101: Task Gate, note that in this case, the Offset value is unused and should be set to zero.
	/// - 0110: 16-bit Interrupt Gate
	/// - 0111: 16-bit Trap Gate
	/// - 1110: 32-bit Interrupt Gate
	/// - 1111: 32-bit Trap Gate
	///
	/// Bit 4: always zero
	///
	/// Bits 5-6: DPL, defines cpu privilege level require to INT this interrupt.
	/// For our purposes, always zero.
	///
	/// Bit 7: 1 if present, 0 if not.
	///
	/// Note that these appear in reverse order in Rust.
	/// For example, 0b1000_1110 sets `present = 1` and `type = 1110`.
	options: u8,

	/// High 16 bits of the 32-bit ISR pointer
	pointer_high: u16,
}

impl<T> fmt::Debug for Entry<T> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("Entry")
			.field("preset", &self.is_present())
			.field(
				"handler_addr",
				&format_args!("{:#x}", self.handler_addr().unwrap_or(VirtAddr::new(0))),
			)
			.field("options", &self.options)
			.finish()
	}
}

// spell:off
impl<F> Entry<F> {
	/// Create a valid non-present IDT entry.
	#[inline]
	pub const fn missing() -> Self {
		Entry {
			pointer_low: 0,
			segment: 0,
			reserved: 0,
			options: 0,
			pointer_high: 0,
			phantom: PhantomData,
		}
	}

	/// Sets the handler address for the IDT entry and sets the following defaults:
	///   - The code selector is the code segment currently active in the CPU
	///   - The present bit is set
	///   - Interrupts are disabled on handler invocation
	///   - The privilege level (DPL) is [`PrivilegeLevel::Ring0`]
	///   - No IST is configured (existing stack will be used)
	///   - This is a 32-bit interrupt gate
	///
	/// # Safety
	///
	/// The caller must ensure that `addr` is the address of a valid interrupt handler function,
	/// and the signature of such a function is correct for the entry type.
	#[inline]
	pub unsafe fn set_handler_addr(&mut self, addr: VirtAddr) {
		let ptr = addr.as_u32();

		self.pointer_low = ptr as u16;
		self.pointer_high = (ptr >> 16) as u16;
		self.options = 0b1000_1110; // Present 32-bit interrupt gate

		// SAFETY: The current CS is a valid, long-mode code segment.
		self.segment = super::util::get_cs();
	}

	/// True if the "present" bit is set, false otherwise.
	pub fn is_present(&self) -> bool {
		return self.options & 0b1000_0000 == 0b1000_0000;
	}

	/// Get the address of this entry's handler.
	pub fn handler_addr(&self) -> Option<VirtAddr> {
		self.is_present().then_some(VirtAddr::new(
			self.pointer_low as u32 + ((self.pointer_high as u32) << 16),
		))
	}
}
// spell:on

// spell:off
impl<F: HandlerFuncType> Entry<F> {
	/// Sets the handler address for the IDT entry and sets the following defaults:
	///   - The code selector is the code segment currently active in the CPU
	///   - The present bit is set
	///   - Interrupts are disabled on handler invocation
	///   - The privilege level (DPL) is [`PrivilegeLevel::Ring0`]
	///   - No IST is configured (existing stack will be used)
	///   - This is a 32-bit interrupt gate
	#[inline]
	pub fn set_handler_fn(&mut self, handler: F) {
		unsafe { self.set_handler_addr(handler.to_virt_addr()) }
	}
}
// spell:on
