use core::arch::asm;

use super::EFlags;

/// Disable interrupts
#[inline]
pub fn cli() {
	unsafe {
		asm!("cli", options(preserves_flags, nostack));
	}
}

/// Enable interrupts
#[inline]
pub fn sti() {
	unsafe {
		asm!("sti", options(preserves_flags, nostack));
	}
}

/// Run a closure with disabled interrupts.
///
/// Run the given closure, disabling interrupts before running it (if they aren't already disabled).
/// Afterwards, interrupts are enabling again if they were enabled before.
///
/// If you have other `enable` and `disable` calls _within_ the closure, things may not work as expected.
#[inline]
pub fn without_interrupts<F, R>(f: F) -> R
where
	F: FnOnce() -> R,
{
	// true if the interrupt flag is set (i.e. interrupts are enabled)
	let saved_intpt_flag = EFlags::read().contains(EFlags::INTERRUPT_FLAG);

	// if interrupts are enabled, disable them for now
	if saved_intpt_flag {
		cli();
	}

	let ret = f();

	if saved_intpt_flag {
		sti();
	}

	ret
}
