use rand::seq::IndexedRandom;

use crate::{drivers::vga::VgaColor, RNG};

#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub enum Tetromino {
	/// The 1x4 line
	Straight,

	/// The 2x2 square
	Square,

	/// The "L" shape
	AngleRight,

	/// The mirror "L" shape
	AngleLeft,

	/// The skew shape "Z"
	SkewRight,

	/// The mirror skew shape
	SkewLeft,

	/// The "T" shape
	Tee,
}

impl Tetromino {
	/// Pick a random tetromino
	pub fn choose_rand() -> Self {
		let ominos = [
			Tetromino::Straight,
			Tetromino::Square,
			Tetromino::AngleRight,
			Tetromino::AngleLeft,
			Tetromino::SkewRight,
			Tetromino::SkewLeft,
			Tetromino::Tee,
		];

		let mut rng = RNG.lock();
		*(ominos.choose(&mut rng).unwrap())
	}
}

#[derive(Debug, Clone, Copy)]
pub enum Direction {
	North,
	East,
	South,
	West,
}

impl Direction {
	pub fn rot_cw(self) -> Self {
		match self {
			Self::North => Self::East,
			Self::East => Self::South,
			Self::South => Self::West,
			Self::West => Self::North,
		}
	}

	pub fn rot_ccw(self) -> Self {
		match self {
			Self::North => Self::West,
			Self::West => Self::South,
			Self::South => Self::East,
			Self::East => Self::North,
		}
	}
}

#[derive(Debug, Clone)]
pub struct FallingTetromino {
	tetromino: Tetromino,
	direction: Direction,
	center_x: usize,
	center_y: usize,

	pub color: VgaColor,
}

impl FallingTetromino {
	pub fn new(tetromino: Tetromino, color: VgaColor, center_x: usize, center_y: usize) -> Self {
		Self {
			tetromino,
			direction: Direction::North,
			color,
			center_x,
			center_y,
		}
	}

	pub fn random(center_x: usize, center_y: usize) -> Self {
		Self::new(
			Tetromino::choose_rand(),
			VgaColor::choose_rand(),
			center_x,
			center_y,
		)
	}

	pub fn translate(&mut self, x: i16, y: i16) {
		if x > 0 {
			let x = usize::try_from(x).unwrap();
			self.center_x += x;
		} else if x < 0 {
			let x = usize::try_from(-x).unwrap();
			self.center_x -= x;
		}

		if y > 0 {
			let y = usize::try_from(y).unwrap();
			self.center_y += y;
		} else if y < 0 {
			let y = usize::try_from(-y).unwrap();
			self.center_y -= y;
		}
	}

	pub fn rotate_cw(&mut self) {
		self.direction = self.direction.rot_cw()
	}

	pub fn rotate_ccw(&mut self) {
		self.direction = self.direction.rot_ccw()
	}

	/// Returns the positions of this falling tetromino's tiles.
	pub fn tiles(&self) -> [(usize, usize); 4] {
		match (&self.tetromino, self.direction) {
			(Tetromino::Square, _) => [
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x + 1, self.center_y + 1),
			],

			//
			// Straight
			//
			(Tetromino::Straight, Direction::North) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x, self.center_y + 2),
			],

			(Tetromino::Straight, Direction::East) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x + 2, self.center_y),
			],

			(Tetromino::Straight, Direction::South) => [
				(self.center_x + 1, self.center_y - 1),
				(self.center_x + 1, self.center_y),
				(self.center_x + 1, self.center_y + 1),
				(self.center_x + 1, self.center_y + 2),
			],

			(Tetromino::Straight, Direction::West) => [
				(self.center_x - 1, self.center_y + 1),
				(self.center_x, self.center_y + 1),
				(self.center_x + 1, self.center_y + 1),
				(self.center_x + 2, self.center_y + 1),
			],

			//
			// Right Angle
			//
			(Tetromino::AngleRight, Direction::North) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x + 1, self.center_y + 1),
			],

			(Tetromino::AngleRight, Direction::East) => [
				(self.center_x - 1, self.center_y + 1),
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
			],

			(Tetromino::AngleRight, Direction::South) => [
				(self.center_x - 1, self.center_y - 1),
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
			],

			(Tetromino::AngleRight, Direction::West) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x + 1, self.center_y - 1),
			],

			//
			// Left Angle
			//
			(Tetromino::AngleLeft, Direction::North) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x - 1, self.center_y + 1),
			],

			(Tetromino::AngleLeft, Direction::East) => [
				(self.center_x - 1, self.center_y - 1),
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
			],

			(Tetromino::AngleLeft, Direction::South) => [
				(self.center_x + 1, self.center_y - 1),
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
			],

			(Tetromino::AngleLeft, Direction::West) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x + 1, self.center_y + 1),
			],

			//
			// Left Skew
			//
			(Tetromino::SkewLeft, Direction::North) => [
				(self.center_x - 1, self.center_y + 1),
				(self.center_x, self.center_y + 1),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
			],

			(Tetromino::SkewLeft, Direction::East) => [
				(self.center_x - 1, self.center_y - 1),
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
			],

			(Tetromino::SkewLeft, Direction::South) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y - 1),
				(self.center_x + 1, self.center_y - 1),
			],

			(Tetromino::SkewLeft, Direction::West) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x + 1, self.center_y + 1),
			],

			//
			// Right Skew
			//
			(Tetromino::SkewRight, Direction::North) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x + 1, self.center_y + 1),
			],

			(Tetromino::SkewRight, Direction::East) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x - 1, self.center_y),
				(self.center_x - 1, self.center_y + 1),
			],

			(Tetromino::SkewRight, Direction::South) => [
				(self.center_x - 1, self.center_y - 1),
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
			],

			(Tetromino::SkewRight, Direction::West) => [
				(self.center_x + 1, self.center_y - 1),
				(self.center_x + 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
			],

			//
			// Tee
			//
			(Tetromino::Tee, Direction::North) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x, self.center_y - 1),
			],

			(Tetromino::Tee, Direction::East) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x + 1, self.center_y),
			],

			(Tetromino::Tee, Direction::South) => [
				(self.center_x - 1, self.center_y),
				(self.center_x, self.center_y),
				(self.center_x + 1, self.center_y),
				(self.center_x, self.center_y + 1),
			],

			(Tetromino::Tee, Direction::West) => [
				(self.center_x, self.center_y - 1),
				(self.center_x, self.center_y),
				(self.center_x, self.center_y + 1),
				(self.center_x - 1, self.center_y),
			],
		}
	}
}
